# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Version according to SemVer versioning system (https://semver.org/)."""


__all__ = [
    "VERSION",
]


_MAJOR = "0"  # Major version to increment in case of incompatible API changes

_MINOR = (
    "0"  # Minor version to increment in case of backward compatible new functionality
)

_PATCH = "1"  # Patch version to increment in case of backward compatible bug fixes

VERSION = f"{_MAJOR}.{_MINOR}.{_PATCH}"
"""The package version."""
